/* 
   l2capsession - L2CAP session handler	
   Copyright (C) 2003 David Libault <david.libault@inventel.fr>
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License version 2 as
   published by the Free Software Foundation;
   
   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
   OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT OF THIRD PARTY RIGHTS.
   IN NO EVENT SHALL THE COPYRIGHT HOLDER(S) AND AUTHOR(S) BE LIABLE FOR ANY CLAIM,
   OR ANY SPECIAL INDIRECT OR CONSEQUENTIAL DAMAGES, OR ANY DAMAGES WHATSOEVER
   RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
   NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE
   USE OR PERFORMANCE OF THIS SOFTWARE.
   
   ALL LIABILITY, INCLUDING LIABILITY FOR INFRINGEMENT OF ANY PATENTS, COPYRIGHTS,
   TRADEMARKS OR OTHER RIGHTS, RELATING TO USE OF THIS SOFTWARE IS DISCLAIMED.
*/

/*
   $Id$
*/

/* 
   David LIBAULT
   Le 29 Janvier 2003
   l2capsession : l2cap session handler
*/

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <fcntl.h>
#include <sys/socket.h>

#include <bluetooth/bluetooth.h>
#include <bluetooth/l2cap.h>

int l2cap_sock = -1;

int do_open(char *args)
{
	struct l2cap_options l2o;
	int psm, olen, mtu;
	char bd_addr_str[128];
	bdaddr_t dst;
	struct sockaddr_l2 sa;
	int result;
	bdaddr_t src_addr = *BDADDR_ANY;

	if (l2cap_sock != -1) {
		printf("l2cap socket already opened...\n");
		return -1;
	}
	sscanf(args, "%s %d %d", bd_addr_str, &psm, &mtu);
	printf("ba : %s, psm %d, mtu %d\n", bd_addr_str, psm, mtu);
	str2ba(bd_addr_str, &dst);
	l2cap_sock = socket(PF_BLUETOOTH, SOCK_SEQPACKET, BTPROTO_L2CAP);
	if (l2cap_sock < 0) {
		printf("Error, can't open l2cap socket to %s\n", args);
		printf("l2cap_sock : %d\n", l2cap_sock);
		return -1;
	}

	// Set local address
	getsockopt(l2cap_sock, SOL_L2CAP, L2CAP_OPTIONS, &l2o, &olen);
	l2o.imtu = l2o.omtu = mtu;
	setsockopt(l2cap_sock, SOL_L2CAP, L2CAP_OPTIONS, &l2o,
		   sizeof(l2o));

	sa.l2_family = AF_BLUETOOTH;
	sa.l2_psm = 0;
	sa.l2_bdaddr = src_addr;
	if (bind(l2cap_sock, (struct sockaddr *) &sa, sizeof(sa))) {
		printf("Error, Couldn't bind socket\n");
		return -1;
	}

	sa.l2_psm = psm;
	sa.l2_bdaddr = dst;
	result = connect(l2cap_sock, (struct sockaddr *) &sa, sizeof(sa));
	if (result == 0) {
		return 0;
	} else {
		printf("Connection failed, result : %d\n", result);
		l2cap_sock = -1;
		return -1;
	}
	return 0;
}

void do_close()
{
	close(l2cap_sock);
	l2cap_sock = -1;
}

int do_sleep(char *args)
{
	int i;

	sscanf(args, "%d", &i);
	printf("Sleep %d seconds\n", i);
	sleep(i);
	return 0;
}

int do_send(char *args)
{
	unsigned char packet[10000];
	unsigned char byte[5];
	int reqsize = 0;
	int n = 0;
	int i;
	int value;


	if (l2cap_sock < 0) {
		printf("Can't send while l2cap socket is not opened\n");
		return -1;
	}

	do {
		sscanf(args, "%s", byte);
		args += strlen(byte) + 1;
		sscanf(byte, "%x", &value);
		packet[reqsize] = value;
		reqsize++;
	} while (strlen(args));

	printf("Sending %d bytes\n", reqsize);
	for (i = 0; i < reqsize; i++)
		printf("0x%2.2X ", packet[i]);
	printf("\n");
	n = send(l2cap_sock, packet, reqsize, 0);
	if (n == -1) {
		printf("Error sending data\n");
		return -1;
	}
	if (n != reqsize) {
		printf("Error sent %d out of %d bytes\n", n, reqsize);
		return -1;
	}
	return 0;
}

int do_wait(char *args)
{
	int rx_timeout;
	fd_set readFds;
	struct timeval timeout;
	int bytesRecvd = 0;
	char rspbuf[256];
	int i;

	if (l2cap_sock < 0) {
		printf("Can't wait while l2cap socket is not opened\n");
		return -1;
	}
	sscanf(args, "%d", &rx_timeout);
	timeout.tv_sec = rx_timeout;
	timeout.tv_usec = 0;
	FD_SET(l2cap_sock, &readFds);
	printf("Waiting for response\n");
	if (0 == select(l2cap_sock + 1, &readFds, NULL, NULL, &timeout)) {
		printf("Client timed out\n");
		return -1;
	}
	bytesRecvd = recv(l2cap_sock, rspbuf, 256, 0);
	printf("Read %d bytes :\n", bytesRecvd);
	for (i = 0; i < bytesRecvd; i++)
		printf("0x%2.2X ", rspbuf[i]);
	printf("\n");
	return 0;
}

char *freadline(FILE * f)
{
	static char buffer[65000];
	int c, i = 0;
	int escape_next = 0;
	int comments = 0;

	c = fgetc(f);

	while ((!feof(f)) && (i < 65000)) {
		switch (c) {
		case '#':
			// hash : comments, or not ?
			if (escape_next) {
				buffer[i++] = c;
				escape_next = 0;
				break;
			}
			comments = 1;
			break;

		case '\\':
			// escaping next char
			if (escape_next) {
				buffer[i++] = c;
				escape_next = 0;
				break;
			}
			escape_next = 1;
			break;

		case '\n':
			buffer[i++] = 0;
			return buffer;
			break;

		case EOF:
			// end of file
			buffer[i++] = 0;
			return buffer;
			break;

		default:
			// normal
			if (comments == 0)
				buffer[i++] = c;
			break;
		}
		c = fgetc(f);
	}
	buffer[i++] = 0;
	return buffer;
}

int handle_input(char *line)
{
	char command[25];

	sscanf(line, "%s", command);
	printf("Got command %s\n", command);
	if (!strncmp(command, "open", 4)) {
		do_open(line + 4);
	} else if (!strncmp(command, "close", 5)) {
		do_close();
	} else if (!strncmp(command, "send", 4)) {
		do_send(line + 4);
	} else if (!strncmp(command, "wait", 4)) {
		do_wait(line + 4);
	} else if (!strncmp(command, "sleep", 5)) {
		do_sleep(line + 5);
	}
	return 0;
}

int main(int argc, char **argv)
{
	FILE *f_input = NULL;
	char *line_buff;

	if (argc < 1) {
		printf("Usage %s <file>\n", argv[0]);
		exit(1);
	}
	
	f_input = fopen(argv[1], "r");
	if (f_input == NULL) {
		printf("Impossible to open file %s\n", argv[1]);
		return -1;
	}
	while (!feof(f_input)) {
		line_buff = freadline(f_input);
		if (strlen(line_buff)) {
			handle_input(line_buff);
		}
	}
	return 0;
}
